Package: weightederm
Type: Package
Title: Weighted Empirical Risk Minimization for Changepoint Regression
Version: 0.1.0
Authors@R: c(
    person("Gabriel", "Arpino", role = c("aut", "cre"),
           email = "arpino.gabriel@gmail.com"))
Description: R interface to the 'weightederm' package for 'Python', which
    provides 'scikit-learn'-style estimators for offline change point
    regression (data segmentation) via weighted empirical risk minimization.
    Supports least-squares, Huber, and logistic losses with fixed or
    cross-validated numbers of change points. Wraps 'Python' via 'reticulate'.
    Arpino and Venkataramanan (2026) <doi:10.48550/arXiv.2604.11746>.
License: Apache License (>= 2.0)
URL: https://github.com/gabrielarpino/weightederm-r
BugReports: https://github.com/gabrielarpino/weightederm-r/issues
Imports: reticulate (>= 1.28)
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
SystemRequirements: Python (>= 3.9), weightederm Python package
NeedsCompilation: no
Packaged: 2026-04-21 21:45:01 UTC; gabrielarpino
Author: Gabriel Arpino [aut, cre]
Maintainer: Gabriel Arpino <arpino.gabriel@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-22 08:20:02 UTC
Built: R 4.5.2; ; 2026-04-22 09:39:42 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(coef,werm_fit)
S3method(predict,werm_fit)
S3method(print,werm_fit)
S3method(summary,werm_fit)
export(weightederm_configure_python)
export(weightederm_examples_available)
export(werm_huber)
export(werm_huber_cv)
export(werm_least_squares)
export(werm_least_squares_cv)
export(werm_logistic)
export(werm_logistic_cv)

# 𐄷 Weighted ERM - fast & accurate change point regression in R

---

<p align="center">
  <img src="man/figures/raw_marginal_plot_5.svg" alt="weightederm" style="width: 100%; max-width: 800px; height: auto;">
</p>

<p align="center">
  <a href="https://arxiv.org/abs/2604.11746"><img src="https://img.shields.io/badge/arXiv-2604.11746-b31b1b.svg" alt="arXiv"></a>
  <a href="https://doi.org/10.5281/zenodo.19472923">
    <img src="https://zenodo.org/badge/1205016507.svg" alt="DOI">
  </a>
  <br>
  <a href="https://CRAN.R-project.org/package=weightederm">
    <img src="https://www.r-pkg.org/badges/version/weightederm" alt="CRAN status">
  </a>
  <a href="https://github.com/gabrielarpino/weightederm-r/actions/workflows/R-CMD-check.yaml">
    <img src="https://github.com/gabrielarpino/weightederm-r/actions/workflows/R-CMD-check.yaml/badge.svg" alt="R-CMD-check">
  </a>
</p>

> R interface to the [weightederm](https://github.com/gabrielarpino/weightederm)
> Python package — changepoint regression via Weighted Empirical Risk Minimization (WERM).

## Overview

`weightederm` detects changepoints in ordered regression data by minimizing a
Weighted Empirical Risk (WERM). The method is described in:

> [Inferring Change Points in Regression via Sample Weighting](https://arxiv.org/abs/2604.11746) by Gabriel Arpino and Ramji Venkataramanan. 

Six estimators are available:

| Loss           | Fixed `num_chgpts`      | Unknown `num_chgpts` via CV  |
|----------------|-------------------------|------------------------------|
| Least squares  | `werm_least_squares()`  | `werm_least_squares_cv()`    |
| Huber          | `werm_huber()`          | `werm_huber_cv()`            |
| Logistic       | `werm_logistic()`       | `werm_logistic_cv()`         |

## Prerequisites

This package calls Python via [reticulate](https://rstudio.github.io/reticulate/).
You need:

1. **Python ≥ 3.9** — install from <https://www.python.org/> or via conda/pyenv.
2. **The `weightederm` Python package:**

```bash
pip install git+https://github.com/gabrielarpino/weightederm.git
```

## Installation

### From CRAN

```r
install.packages("weightederm")
```

### Development version from GitHub

```r
# install.packages("pak")
pak::pak("gabrielarpino/weightederm-r")
```

Or with `remotes`:

```r
# install.packages("remotes")
remotes::install_github("gabrielarpino/weightederm-r")
```

## Pointing to your Python environment

By default, `reticulate` auto-discovers Python. If it picks the wrong one, call
`weightederm_configure_python()` **before** your first model fit:

```r
# Option 1 — point to a specific Python binary
weightederm_configure_python(python = "/path/to/your/venv/bin/python")

# Option 2 — point to the Python binary inside a conda environment
weightederm_configure_python("/path/to/miniconda/envs/my_conda_env/bin/python")
```

You can also set the `RETICULATE_PYTHON` environment variable in your
`.Renviron` file to avoid calling this every session:

```
RETICULATE_PYTHON=/path/to/your/venv/bin/python
```

## Quick start

### M1 — fixed number of changepoints, sparse linear regression

```r
library(weightederm)
set.seed(0)

n <- 120L; p <- 20L; true_cp <- 60L
X <- matrix(rnorm(n * p), n, p)

beta_left  <- rep(0, p); beta_left[c(1, 4)]  <- c( 2.0, -1.5)
beta_right <- rep(0, p); beta_right[c(1, 4)] <- c(-1.0,  2.5)

y <- numeric(n)
y[1:true_cp]          <- X[1:true_cp, ]          %*% beta_left  + rnorm(true_cp, sd = 0.2)
y[(true_cp+1):n]      <- X[(true_cp+1):n, ]      %*% beta_right + rnorm(n - true_cp, sd = 0.2)

fit <- werm_least_squares(X, y, num_chgpts = 1L, delta = 5L,
                          search_method = "efficient", fit_intercept = FALSE)
fit
#> WERM Changepoint Estimator (werm_least_squares)
#>   Changepoints (1-indexed): 60
#>   num_chgpts: 1 | num_signals: 2 | n_features_in: 20
#>   Objective: ...
```

### M2 — unknown number of changepoints via cross-validation

```r
set.seed(1)
n <- 180L; p <- 10L
X <- matrix(rnorm(n * p), n, p)

beta_1 <- rep(0, p); beta_1[1]      <- 3.5
beta_2 <- rep(0, p); beta_2[1:2]    <- c(-3.0, 3.0)
beta_3 <- rep(0, p); beta_3[1:3]    <- c(2.5, -2.5, 2.5)

y <- numeric(n)
y[1:60]    <- X[1:60, ]    %*% beta_1 + rnorm(60, sd = 0.05)
y[61:120]  <- X[61:120, ]  %*% beta_2 + rnorm(60, sd = 0.05)
y[121:180] <- X[121:180, ] %*% beta_3 + rnorm(60, sd = 0.05)

fit <- werm_least_squares_cv(X, y, max_num_chgpts = 2L, delta = 5L, cv = 3L,
                              search_method = "efficient", fit_intercept = FALSE)
fit$best_num_chgpts  # 2
fit$changepoints     # near c(60, 120)
fit$cv_results       # data.frame of num_chgpts vs mean_test_score
```

### M3 — binary classification with a structural break

```r
set.seed(2)
n <- 160L; p <- 12L; true_cp <- 80L
X <- matrix(rnorm(n * p), n, p)

beta_left  <- rep(0, p); beta_left[c(1, 3)]  <- c( 2.5, -2.0)
beta_right <- rep(0, p); beta_right[c(1, 3)] <- c(-2.5,  2.0)

eta <- numeric(n)
eta[1:true_cp]       <- X[1:true_cp, ]       %*% beta_left
eta[(true_cp+1):n]   <- X[(true_cp+1):n, ]   %*% beta_right
y <- rbinom(n, 1L, 1 / (1 + exp(-eta)))

fit <- werm_logistic(X, y, num_chgpts = 1L, delta = 5L,
                     search_method = "efficient", fit_intercept = FALSE,
                     max_iter = 300L)
fit$changepoints   # near 80
fit$classes        # c("0", "1")

# Predict probabilities on new data
X_new <- matrix(rnorm(10 * p), 10, p)
predict(fit, X_new, type = "prob")   # 10 × 2 matrix
predict(fit, X_new, type = "class")  # character vector of labels
```

## Key arguments

| Argument           | Description                                                          | Default   |
|--------------------|----------------------------------------------------------------------|-----------|
| `num_chgpts`       | Number of changepoints (fixed estimators)                           | required  |
| `max_num_chgpts`   | Upper bound for CV search (CV estimators)                           | required  |
| `delta`            | Minimum gap between changepoints during search                      | `1L`      |
| `search_method`    | `"efficient"` (greedy + local refinement) or `"brute_force"`        | `"efficient"` |
| `fit_intercept`    | Include per-segment intercept                                        | `TRUE`    |
| `cv`               | Number of interleaved CV folds (CV estimators)                       | `5L`      |
| `penalty`          | `"none"`, `"l1"`, or `"l2"`                                         | `"none"` (`"l2"` for logistic) |
| `alpha`            | Penalty strength                                                     | `0.0`     |
| `epsilon`          | Huber transition parameter                                           | `1.35`    |
| `max_iter`         | Optimizer iterations (Huber / logistic)                             | `100L`    |

## Fitted object

Every `werm_*()` call returns a named list with class `c("werm_<type>", "werm_fit")`:

| Element                 | Description                                                |
|-------------------------|------------------------------------------------------------|
| `changepoints`          | Integer vector, **1-indexed** (R convention)               |
| `num_chgpts`            | Number of detected changepoints                            |
| `num_signals`           | `num_chgpts + 1`                                           |
| `objective`             | Minimised WERM value                                       |
| `signal_coefs`          | `(num_signals × p)` Stage-1 WERM coefficient matrix       |
| `signal_intercepts`     | Numeric vector or `NULL`                                   |
| `last_segment_coef`     | Coefficients used by `predict()`                           |
| `last_segment_intercept`| Numeric or `NULL`                                          |
| `n_features_in`         | Number of input features                                   |
| `classes`               | (logistic only) Character vector of length 2               |

CV estimators additionally expose `best_num_chgpts`, `best_score`, `cv_results`
(a data frame), `num_chgpts_grid`, `segment_bounds`, `segment_coefs`, and
`segment_intercepts`.

## S3 methods

```r
print(fit)           # compact summary
summary(fit)         # print + coefficient table
coef(fit)            # last_segment_coef vector
predict(fit, X_new)  # numeric predictions (regression) or class labels / probabilities (logistic)
```

## Related resources

- [User Guide](https://github.com/gabrielarpino/weightederm-r/blob/main/docs/user_guide.md) — algorithm overview, fixed vs CV, penalties, prediction, S3 methods (with tested R examples)
- [weightederm Python package](https://github.com/gabrielarpino/weightederm) — Python API, parameter reference, benchmark notebook

## License

Apache License 2.0 — the same license as the underlying Python package.

# ---------------------------------------------------------------------------
# Fixed-changepoint estimators
# ---------------------------------------------------------------------------

#' Fit a WERM changepoint model with squared loss
#'
#' Detects `num_chgpts` changepoints in ordered regression data by minimising
#' a Weighted Empirical Risk with squared loss.
#'
#' @param X Numeric matrix of shape `(n, p)`. Observations must be in order.
#' @param y Numeric vector of length `n`. Ordered responses.
#' @param num_chgpts Integer. Number of changepoints to detect.
#' @param delta Integer. Minimum gap between candidate changepoints during
#'   search. Default `1`. Rule of thumb: `max(1L, as.integer(0.05 * nrow(X)))`.
#' @param search_method Character. `"efficient"` (default) or `"brute_force"`.
#' @param fit_intercept Logical. Whether each segment model includes an
#'   intercept. Default `TRUE`.
#' @param fit_solver Character. `"direct"` (default, closed-form) or
#'   `"lbfgsb"` (gradient-based).
#' @param penalty Character. `"none"` (default), `"l1"`, or `"l2"`.
#' @param alpha Numeric. Penalty strength. Default `0`.
#'
#' @return An object of class `c("werm_least_squares", "werm_fit")` with the
#'   following named elements:
#'   \describe{
#'     \item{`changepoints`}{Integer vector of detected changepoints
#'       (**1-indexed**, R convention). Length equals `num_chgpts`.}
#'     \item{`num_chgpts`}{Integer. Number of detected changepoints.}
#'     \item{`num_signals`}{Integer. Number of segments (`num_chgpts + 1`).}
#'     \item{`objective`}{Numeric. Minimised WERM objective value.}
#'     \item{`last_segment_coef`}{Numeric vector. Coefficient of the
#'       unweighted base-loss refit on the last segment (used by `predict()`).}
#'     \item{`last_segment_intercept`}{Numeric or `NULL`.}
#'     \item{`signal_coefs`}{Matrix `(num_signals x p)`. Stage-1 WERM
#'       coefficient estimates.}
#'     \item{`signal_intercepts`}{Numeric vector or `NULL`.}
#'     \item{`n_features_in`}{Integer. Number of features.}
#'   }
#'
#' @examples
#' # Limit BLAS/OpenMP threads so example CPU time stays proportional to
#' # elapsed time on multicore CRAN machines.
#' Sys.setenv(
#'   OMP_NUM_THREADS = "1",
#'   OPENBLAS_NUM_THREADS = "1",
#'   MKL_NUM_THREADS = "1",
#'   BLAS_NUM_THREADS = "1"
#' )
#'
#' if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
#'     weightederm_examples_available("WERMLeastSquares")) {
#'   set.seed(1)
#'   n <- 24L; p <- 2L; true_cp <- 12L
#'   X <- matrix(rnorm(n * p), n, p)
#'   y <- c(
#'     X[1:true_cp, ] %*% c(3, -1.5),
#'     X[(true_cp + 1L):n, ] %*% c(-3, 1.5)
#'   ) + rnorm(n, sd = 0.05)
#'   fit <- werm_least_squares(X, y, num_chgpts = 1L, delta = 3L,
#'                             fit_intercept = FALSE)
#'   fit$changepoints
#' }
#'
#' @export
werm_least_squares <- function(
    X, y,
    num_chgpts,
    delta          = 1L,
    search_method  = "efficient",
    fit_intercept  = TRUE,
    fit_solver     = "direct",
    penalty        = "none",
    alpha          = 0.0
) {
  py <- .weightederm_py()
  X_py <- .to_numpy_matrix(X)
  y_py <- .to_numpy_vector(y)

  py_model <- py$WERMLeastSquares(
    num_chgpts    = as.integer(num_chgpts),
    delta         = as.integer(delta),
    search_method = search_method,
    fit_intercept = fit_intercept,
    fit_solver    = fit_solver,
    penalty       = penalty,
    alpha         = as.numeric(alpha)
  )
  py_model$fit(X_py, y_py)

  attrs <- .extract_fixed_attrs(py_model)
  .build_werm_object(attrs, "least_squares", py_model)
}


#' Fit a WERM changepoint model with Huber loss
#'
#' Like [werm_least_squares()] but uses the Huber loss, which is more robust
#' to outliers than squared loss.
#'
#' @inheritParams werm_least_squares
#' @param epsilon Numeric. Huber transition parameter. Default `1.35`
#'   (95 % Gaussian efficiency).
#' @param max_iter Integer. Maximum L-BFGS-B iterations. Default `100`.
#' @param tol Numeric. Gradient-norm tolerance. Default `1e-5`.
#'
#' @return An object of class `c("werm_huber", "werm_fit")`.
#'   Same elements as [werm_least_squares()].
#'
#' @examples
#' # Limit BLAS/OpenMP threads so example CPU time stays proportional to
#' # elapsed time on multicore CRAN machines.
#' Sys.setenv(
#'   OMP_NUM_THREADS = "1",
#'   OPENBLAS_NUM_THREADS = "1",
#'   MKL_NUM_THREADS = "1",
#'   BLAS_NUM_THREADS = "1"
#' )
#'
#' if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
#'     weightederm_examples_available("WERMHuber")) {
#'   set.seed(2)
#'   n <- 12L; p <- 1L; true_cp <- 6L
#'   X <- matrix(rnorm(n * p), n, p)
#'   y <- c(
#'     X[1:true_cp, , drop = FALSE] %*% 3,
#'     X[(true_cp + 1L):n, , drop = FALSE] %*% -3
#'   ) + rnorm(n, sd = 0.03)
#'   fit <- werm_huber(X, y, num_chgpts = 1L, delta = 1L,
#'                     fit_intercept = FALSE, max_iter = 5L, tol = 1e-3)
#'   fit$changepoints
#' }
#'
#' @export
werm_huber <- function(
    X, y,
    num_chgpts,
    delta         = 1L,
    search_method = "efficient",
    fit_intercept = TRUE,
    epsilon       = 1.35,
    max_iter      = 100L,
    tol           = 1e-5,
    penalty       = "none",
    alpha         = 0.0
) {
  py <- .weightederm_py()
  X_py <- .to_numpy_matrix(X)
  y_py <- .to_numpy_vector(y)

  py_model <- py$WERMHuber(
    num_chgpts    = as.integer(num_chgpts),
    delta         = as.integer(delta),
    search_method = search_method,
    fit_intercept = fit_intercept,
    epsilon       = as.numeric(epsilon),
    max_iter      = as.integer(max_iter),
    tol           = as.numeric(tol),
    penalty       = penalty,
    alpha         = as.numeric(alpha)
  )
  py_model$fit(X_py, y_py)

  attrs <- .extract_fixed_attrs(py_model)
  .build_werm_object(attrs, "huber", py_model)
}


#' Fit a WERM changepoint model with binary logistic loss
#'
#' Detects `num_chgpts` changepoints in ordered binary classification data.
#'
#' @inheritParams werm_least_squares
#' @param y Integer or factor vector of binary labels (two unique values).
#' @param max_iter Integer. Maximum L-BFGS-B iterations. Default `100`.
#' @param tol Numeric. Gradient-norm tolerance. Default `1e-5`.
#' @param penalty Character. Default `"l2"` (prevents divergence on separable
#'   segments). Set `"none"` only when segments are not separable.
#' @param alpha Numeric. Default `1.0`.
#'
#' @return An object of class `c("werm_logistic", "werm_fit")`.
#'   Contains all elements from [werm_least_squares()] plus:
#'   \describe{
#'     \item{`classes`}{Character vector of length 2. The two class labels in
#'       sorted order. `classes[2]` is the positive class.}
#'   }
#'
#' @examples
#' # Limit BLAS/OpenMP threads so example CPU time stays proportional to
#' # elapsed time on multicore CRAN machines.
#' Sys.setenv(
#'   OMP_NUM_THREADS = "1",
#'   OPENBLAS_NUM_THREADS = "1",
#'   MKL_NUM_THREADS = "1",
#'   BLAS_NUM_THREADS = "1"
#' )
#'
#' if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
#'     weightederm_examples_available("WERMLogistic")) {
#'   set.seed(3)
#'   n <- 30L; p <- 2L; true_cp <- 15L
#'   X <- matrix(rnorm(n * p), n, p)
#'   eta <- c(
#'     X[1:true_cp, ] %*% c(3, -3),
#'     X[(true_cp + 1L):n, ] %*% c(-3, 3)
#'   )
#'   y <- rbinom(n, 1L, 1 / (1 + exp(-eta)))
#'   fit <- werm_logistic(X, y, num_chgpts = 1L, delta = 3L,
#'                        fit_intercept = FALSE, max_iter = 100L)
#'   fit$changepoints
#'   fit$classes
#' }
#'
#' @export
werm_logistic <- function(
    X, y,
    num_chgpts,
    delta         = 1L,
    search_method = "efficient",
    fit_intercept = TRUE,
    max_iter      = 100L,
    tol           = 1e-5,
    penalty       = "l2",
    alpha         = 1.0
) {
  py <- .weightederm_py()
  X_py <- .to_numpy_matrix(X)
  y_py <- .to_numpy_vector(y)

  py_model <- py$WERMLogistic(
    num_chgpts    = as.integer(num_chgpts),
    delta         = as.integer(delta),
    search_method = search_method,
    fit_intercept = fit_intercept,
    max_iter      = as.integer(max_iter),
    tol           = as.numeric(tol),
    penalty       = penalty,
    alpha         = as.numeric(alpha)
  )
  py_model$fit(X_py, y_py)

  attrs <- .extract_fixed_attrs(py_model)
  attrs$classes <- as.character(reticulate::py_to_r(py_model$classes_))

  .build_werm_object(attrs, "logistic", py_model)
}

skip_if_not(weightederm_available, "weightederm Python package not available")

## ── Shared data ─────────────────────────────────────────────────────────────

make_ls_data <- function(seed = 1L, n = 80L, p = 4L, true_cp = 40L) {
  set.seed(seed)
  X <- matrix(rnorm(n * p), nrow = n, ncol = p)
  beta_left  <- c(2, -1, 0, 0)
  beta_right <- c(-1, 2, 0, 0)
  y <- c(
    X[seq_len(true_cp), ] %*% beta_left  + rnorm(true_cp,  sd = 0.1),
    X[seq(true_cp + 1L, n), ] %*% beta_right + rnorm(n - true_cp, sd = 0.1)
  )
  list(X = X, y = y, true_cp = true_cp, n = n, p = p)
}

## ── Basic fit ────────────────────────────────────────────────────────────────

test_that("werm_least_squares returns werm_fit with correct classes", {
  d <- make_ls_data()
  fit <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                             fit_intercept = FALSE)

  expect_s3_class(fit, "werm_fit")
  expect_s3_class(fit, "werm_least_squares")
})

test_that("werm_least_squares detects the correct changepoint", {
  d <- make_ls_data()
  fit <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                             fit_intercept = FALSE)

  expect_equal(fit$num_chgpts, 1L)
  expect_length(fit$changepoints, 1L)
  expect_lte(abs(fit$changepoints[1] - d$true_cp), 5L)
})

test_that("changepoints are 1-indexed", {
  d <- make_ls_data()
  fit <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                             fit_intercept = FALSE)

  expect_gte(fit$changepoints[1], 1L)
  expect_lte(fit$changepoints[1], d$n)
})

test_that("fitted attrs have correct shapes", {
  d <- make_ls_data()
  fit <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                             fit_intercept = FALSE)

  expect_equal(fit$num_signals, 2L)
  expect_length(fit$last_segment_coef, d$p)
  expect_null(fit$last_segment_intercept)        # fit_intercept = FALSE
  expect_equal(dim(fit$signal_coefs), c(2L, d$p))
  expect_equal(fit$n_features_in, d$p)
})

test_that("objective is finite", {
  d <- make_ls_data()
  fit <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                             fit_intercept = FALSE)

  expect_true(is.finite(fit$objective))
})

## ── predict ──────────────────────────────────────────────────────────────────

test_that("predict returns numeric vector of correct length", {
  d <- make_ls_data()
  fit <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                             fit_intercept = FALSE)
  X_new <- matrix(rnorm(10L * d$p), nrow = 10L)
  preds <- predict(fit, X_new)

  expect_type(preds, "double")
  expect_length(preds, 10L)
  expect_true(all(is.finite(preds)))
})

## ── coef ─────────────────────────────────────────────────────────────────────

test_that("coef() returns last_segment_coef", {
  d <- make_ls_data()
  fit <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                             fit_intercept = FALSE)

  expect_equal(coef(fit), fit$last_segment_coef)
})

## ── num_chgpts = 0 ───────────────────────────────────────────────────────────

test_that("werm_least_squares works with zero changepoints", {
  set.seed(99)
  n <- 40L; p <- 2L
  X <- matrix(rnorm(n * p), n, p)
  y <- X %*% c(1, -1) + rnorm(n, sd = 0.1)
  fit <- werm_least_squares(X, y, num_chgpts = 0L, delta = 3L,
                             fit_intercept = FALSE)

  expect_equal(fit$num_chgpts, 0L)
  expect_length(fit$changepoints, 0L)
})

## ── L2 penalty ───────────────────────────────────────────────────────────────

test_that("L2 penalty shrinks last_segment_coef toward zero", {
  d <- make_ls_data()
  fit_none <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                                  fit_intercept = FALSE, penalty = "none")
  fit_l2   <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                                  fit_intercept = FALSE, penalty = "l2",
                                  alpha = 5.0)

  expect_lte(sqrt(sum(fit_l2$last_segment_coef^2)),
             sqrt(sum(fit_none$last_segment_coef^2)) + 1e-9)
})

## ── print / summary smoke test ───────────────────────────────────────────────

test_that("print.werm_fit runs without error", {
  d <- make_ls_data()
  fit <- werm_least_squares(d$X, d$y, num_chgpts = 1L, delta = 5L,
                             fit_intercept = FALSE)
  expect_output(print(fit), "WERM Changepoint Estimator")
  expect_output(print(fit), "1-indexed")
})

skip_if_not(weightederm_available, "weightederm Python package not available")

## R translations of the M1, M2, M3 minimal working examples from the Python
## README. Data generation uses the same structure and signal strength as the
## Python examples but with R's RNG (different seeds → different draws, same
## detection quality).

## ── M1-style: fixed num_chgpts, sparse linear regression ─────────────────────

test_that("M1: werm_least_squares detects changepoint near position 60", {
  set.seed(0L)
  n <- 120L; p <- 20L; true_cp <- 60L
  X <- matrix(rnorm(n * p), n, p)

  beta_left  <- rep(0, p); beta_left[c(1L, 4L)]  <- c(2.0, -1.5)
  beta_right <- rep(0, p); beta_right[c(1L, 4L)] <- c(-1.0, 2.5)

  y <- numeric(n)
  y[seq_len(true_cp)]       <- X[seq_len(true_cp), ]       %*% beta_left  +
    rnorm(true_cp,      sd = 0.2)
  y[seq(true_cp + 1L, n)] <- X[seq(true_cp + 1L, n), ] %*% beta_right +
    rnorm(n - true_cp,  sd = 0.2)

  fit <- werm_least_squares(X, y, num_chgpts = 1L, delta = 5L,
                             search_method = "efficient",
                             fit_intercept = FALSE)

  expect_equal(fit$num_chgpts, 1L)
  expect_lte(abs(fit$changepoints[1] - true_cp), 5L)
})

## ── M2-style: unknown num_chgpts, CV selection ────────────────────────────────

test_that("M2: werm_least_squares_cv selects 2 changepoints near positions 60 and 120", {
  set.seed(1L)
  n <- 180L; p <- 10L
  true_cps <- c(60L, 120L)
  X <- matrix(rnorm(n * p), n, p)

  beta_1 <- rep(0, p); beta_1[1L]        <- 3.5
  beta_2 <- rep(0, p); beta_2[c(1L, 2L)] <- c(-3.0, 3.0)
  beta_3 <- rep(0, p); beta_3[1L:3L]     <- c(2.5, -2.5, 2.5)

  y <- numeric(n)
  y[1L:60L]    <- X[1L:60L,    ] %*% beta_1 + rnorm(60L, sd = 0.05)
  y[61L:120L]  <- X[61L:120L,  ] %*% beta_2 + rnorm(60L, sd = 0.05)
  y[121L:180L] <- X[121L:180L, ] %*% beta_3 + rnorm(60L, sd = 0.05)

  fit <- werm_least_squares_cv(X, y, max_num_chgpts = 2L, delta = 5L, cv = 3L,
                                search_method = "efficient",
                                fit_intercept = FALSE)

  expect_equal(fit$best_num_chgpts, 2L)
  expect_lte(max(abs(sort(fit$changepoints) - true_cps)), 5L)
})

## ── M3-style: fixed num_chgpts, binary logistic regression ───────────────────

test_that("M3: werm_logistic detects changepoint near position 80", {
  set.seed(2L)
  n <- 160L; p <- 12L; true_cp <- 80L
  X <- matrix(rnorm(n * p), n, p)

  beta_left  <- rep(0, p); beta_left[c(1L, 3L)]  <- c(2.5, -2.0)
  beta_right <- rep(0, p); beta_right[c(1L, 3L)] <- c(-2.5, 2.0)

  eta <- numeric(n)
  eta[seq_len(true_cp)]       <- X[seq_len(true_cp), ]       %*% beta_left
  eta[seq(true_cp + 1L, n)] <- X[seq(true_cp + 1L, n), ] %*% beta_right
  prob <- 1.0 / (1.0 + exp(-eta))
  y <- rbinom(n, 1L, prob)

  fit <- werm_logistic(X, y, num_chgpts = 1L, delta = 5L,
                       search_method = "efficient", fit_intercept = FALSE,
                       max_iter = 300L, tol = 1e-6)

  expect_s3_class(fit, "werm_fit")
  expect_equal(fit$num_chgpts, 1L)
  expect_lte(abs(fit$changepoints[1] - true_cp), 5L)
})

skip_if_not(weightederm_available, "weightederm Python package not available")

make_logistic_data <- function(seed = 3L, n = 120L, p = 4L, true_cp = 60L) {
  set.seed(seed)
  X <- matrix(rnorm(n * p), nrow = n, ncol = p)
  eta <- c(
    X[seq_len(true_cp), ]      %*% c(2.5, -2.0, 0, 0),
    X[seq(true_cp + 1L, n), ]  %*% c(-2.5, 2.0, 0, 0)
  )
  y <- rbinom(n, 1L, 1 / (1 + exp(-eta)))
  list(X = X, y = y, true_cp = true_cp, n = n, p = p)
}

## ── Basic fit ────────────────────────────────────────────────────────────────

test_that("werm_logistic returns werm_logistic / werm_fit", {
  d <- make_logistic_data()
  fit <- werm_logistic(d$X, d$y, num_chgpts = 1L, delta = 5L,
                       fit_intercept = FALSE)

  expect_s3_class(fit, "werm_fit")
  expect_s3_class(fit, "werm_logistic")
})

test_that("werm_logistic detects correct changepoint", {
  d <- make_logistic_data()
  fit <- werm_logistic(d$X, d$y, num_chgpts = 1L, delta = 5L,
                       fit_intercept = FALSE, max_iter = 300L)

  expect_lte(abs(fit$changepoints[1] - d$true_cp), 6L)
})

test_that("classes attribute has length 2", {
  d <- make_logistic_data()
  fit <- werm_logistic(d$X, d$y, num_chgpts = 1L, delta = 5L,
                       fit_intercept = FALSE)

  expect_length(fit$classes, 2L)
  expect_type(fit$classes, "character")
})

## ── predict ──────────────────────────────────────────────────────────────────

test_that("predict(type='class') returns character labels", {
  d <- make_logistic_data()
  fit <- werm_logistic(d$X, d$y, num_chgpts = 1L, delta = 5L,
                       fit_intercept = FALSE)
  X_new <- matrix(rnorm(20L * d$p), 20L, d$p)
  preds <- predict(fit, X_new, type = "class")

  expect_type(preds, "character")
  expect_length(preds, 20L)
  expect_true(all(preds %in% fit$classes))
})

test_that("predict(type='prob') returns matrix with two columns summing to 1", {
  d <- make_logistic_data()
  fit <- werm_logistic(d$X, d$y, num_chgpts = 1L, delta = 5L,
                       fit_intercept = FALSE)
  X_new <- matrix(rnorm(15L * d$p), 15L, d$p)
  probs <- predict(fit, X_new, type = "prob")

  expect_equal(ncol(probs), 2L)
  expect_equal(nrow(probs), 15L)
  expect_true(all(abs(rowSums(probs) - 1) < 1e-9))
  expect_true(all(probs >= 0 & probs <= 1))
})

skip_if_not(weightederm_available, "weightederm Python package not available")

make_huber_data <- function(seed = 5L, n = 80L, p = 3L, true_cp = 40L) {
  set.seed(seed)
  X <- matrix(rnorm(n * p), nrow = n, ncol = p)
  y <- c(
    X[seq_len(true_cp), ]     %*% c(2, -1, 0),
    X[seq(true_cp + 1L, n), ] %*% c(-2, 1, 0)
  ) + rnorm(n, sd = 0.1)
  list(X = X, y = y, true_cp = true_cp, n = n, p = p)
}

## ── Basic fit ────────────────────────────────────────────────────────────────

test_that("werm_huber returns werm_huber / werm_fit", {
  d <- make_huber_data()
  fit <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                   fit_intercept = FALSE)

  expect_s3_class(fit, "werm_fit")
  expect_s3_class(fit, "werm_huber")
})

test_that("werm_huber detects the correct changepoint", {
  d <- make_huber_data()
  fit <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                   fit_intercept = FALSE, max_iter = 200L)

  expect_equal(fit$num_chgpts, 1L)
  expect_length(fit$changepoints, 1L)
  expect_lte(abs(fit$changepoints[1] - d$true_cp), 5L)
})

test_that("changepoints are 1-indexed", {
  d <- make_huber_data()
  fit <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                   fit_intercept = FALSE)

  expect_gte(fit$changepoints[1], 1L)
  expect_lte(fit$changepoints[1], d$n)
})

test_that("fitted attrs have correct shapes", {
  d <- make_huber_data()
  fit <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                   fit_intercept = FALSE)

  expect_equal(fit$num_signals, 2L)
  expect_length(fit$last_segment_coef, d$p)
  expect_null(fit$last_segment_intercept)        # fit_intercept = FALSE
  expect_equal(dim(fit$signal_coefs), c(2L, d$p))
  expect_equal(fit$n_features_in, d$p)
})

test_that("objective is finite", {
  d <- make_huber_data()
  fit <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                   fit_intercept = FALSE)

  expect_true(is.finite(fit$objective))
})

## ── predict ──────────────────────────────────────────────────────────────────

test_that("predict returns numeric vector of correct length", {
  d <- make_huber_data()
  fit <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                   fit_intercept = FALSE)
  X_new <- matrix(rnorm(10L * d$p), nrow = 10L)
  preds <- predict(fit, X_new)

  expect_type(preds, "double")
  expect_length(preds, 10L)
  expect_true(all(is.finite(preds)))
})

## ── num_chgpts = 0 ───────────────────────────────────────────────────────────

test_that("werm_huber works with zero changepoints", {
  set.seed(88L)
  n <- 40L; p <- 2L
  X <- matrix(rnorm(n * p), n, p)
  y <- X %*% c(1, -1) + rnorm(n, sd = 0.1)
  fit <- werm_huber(X, y, num_chgpts = 0L, delta = 3L, fit_intercept = FALSE)

  expect_equal(fit$num_chgpts, 0L)
  expect_length(fit$changepoints, 0L)
})

## ── epsilon parameter ────────────────────────────────────────────────────────

test_that("epsilon parameter is accepted and both fits are valid", {
  d <- make_huber_data()
  fit_small <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                          fit_intercept = FALSE, epsilon = 1.0)
  fit_large <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                          fit_intercept = FALSE, epsilon = 50.0)

  expect_s3_class(fit_small, "werm_fit")
  expect_s3_class(fit_large, "werm_fit")
  expect_true(is.finite(fit_small$objective))
  expect_true(is.finite(fit_large$objective))
})

## ── L2 penalty ───────────────────────────────────────────────────────────────

test_that("L2 penalty shrinks last_segment_coef toward zero", {
  d <- make_huber_data()
  fit_none <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                         fit_intercept = FALSE, penalty = "none")
  fit_l2   <- werm_huber(d$X, d$y, num_chgpts = 1L, delta = 5L,
                         fit_intercept = FALSE, penalty = "l2", alpha = 5.0)

  expect_lte(sqrt(sum(fit_l2$last_segment_coef^2)),
             sqrt(sum(fit_none$last_segment_coef^2)) + 1e-9)
})

## ── robustness to outliers ───────────────────────────────────────────────────

test_that("werm_huber detects changepoint despite large outliers", {
  set.seed(42L)
  n <- 80L; p <- 2L; true_cp <- 40L
  X <- matrix(rnorm(n * p), n, p)
  y <- c(
    X[seq_len(true_cp), ]     %*% c(3, 0),
    X[seq(true_cp + 1L, n), ] %*% c(-3, 0)
  ) + rnorm(n, sd = 0.1)
  # Inject large outliers
  y[c(5L, 15L, 55L, 65L, 75L)] <- y[c(5L, 15L, 55L, 65L, 75L)] + 20

  fit <- werm_huber(X, y, num_chgpts = 1L, delta = 5L,
                   fit_intercept = FALSE, max_iter = 200L)

  expect_lte(abs(fit$changepoints[1] - true_cp), 8L)
})

# Internal module handle — populated on first use via .weightederm_py()
.werm_env <- new.env(parent = emptyenv())

.onLoad <- function(libname, pkgname) {
  reticulate::configure_environment(pkgname)
}

# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------

#' @noRd
.weightederm_py <- function() {
  if (is.null(.werm_env$module)) {
    tryCatch(
      {
        .werm_env$module <- reticulate::import("weightederm", delay_load = FALSE)
      },
      error = function(e) {
        stop(
          "The Python 'weightederm' package could not be imported.\n",
          "Install it with:\n",
          "  pip install weightederm\n",
          "or, if you use the development source:\n",
          "  pip install -e /path/to/weightederm\n\n",
          "To point reticulate at a specific Python / virtual environment call\n",
          "  reticulate::use_python('/path/to/python')\n",
          "before loading this package.\n\n",
          "Original error: ", conditionMessage(e),
          call. = FALSE
        )
      }
    )
  }
  .werm_env$module
}

#' Configure which Python environment weightederm uses
#'
#' A thin wrapper around [reticulate::use_python()] that must be called
#' **before** any `werm_*` function if the default Python does not have
#' `weightederm` installed.
#'
#' @param python Path to the Python binary or virtual-environment directory.
#' @param required Passed to [reticulate::use_python()].
#'
#' @return Invisibly returns \code{NULL}. Called for side effects only.
#'
#' @examples
#' if (nzchar(Sys.which("python"))) {
#'   weightederm_configure_python(Sys.which("python"), required = FALSE)
#' }
#'
#' @export
weightederm_configure_python <- function(python, required = TRUE) {
  .werm_env$module <- NULL  # reset cached module
  reticulate::use_python(python, required = required)
  invisible(NULL)
}

skip_if_not(weightederm_available, "weightederm Python package not available")

## Extract every ```r block whose first line is "# R_DOCS_TEST: <tag>" and
## return a named list of code strings.
.extract_r_docs_blocks <- function(md_file) {
  lines    <- readLines(md_file, warn = FALSE)
  blocks   <- list()
  in_block <- FALSE
  tag      <- NULL
  buf      <- character(0)

  for (line in lines) {
    if (!in_block) {
      if (grepl("^```[rR]\\s*$", line)) {
        in_block <- TRUE
        tag      <- NULL
        buf      <- character(0)
      }
    } else if (grepl("^```\\s*$", line)) {
      if (!is.null(tag)) blocks[[tag]] <- paste(buf, collapse = "\n")
      in_block <- FALSE
    } else {
      if (is.null(tag)) {
        m <- regmatches(line, regexpr("^#\\s*R_DOCS_TEST:\\s*(\\S+)", line, perl = TRUE))
        if (length(m) > 0L) {
          tag <- sub("^#\\s*R_DOCS_TEST:\\s*", "", m)
        } else {
          in_block <- FALSE   # untagged block — ignore
        }
      } else {
        buf <- c(buf, line)
      }
    }
  }

  blocks
}

## ── Run all tagged blocks ────────────────────────────────────────────────────

test_that("all R_DOCS_TEST blocks in user_guide.md execute without error", {
  md_file <- file.path(
    dirname(dirname(getwd())),   # tests/testthat → tests → weightederm-r
    "docs", "user_guide.md"
  )
  skip_if_not(file.exists(md_file), "docs/user_guide.md not found")

  blocks <- .extract_r_docs_blocks(md_file)
  expect_gt(length(blocks), 0L)   # sanity: we actually found some

  for (tag in names(blocks)) {
    env <- new.env(parent = globalenv())
    result <- tryCatch(
      eval(parse(text = blocks[[tag]]), envir = env),
      error = function(e) e
    )
    expect_false(
      inherits(result, "error"),
      label = sprintf("docs block '%s' must run without error (got: %s)",
                      tag,
                      if (inherits(result, "error")) conditionMessage(result) else "ok")
    )
  }
})

## ── Individual block smoke-tests (named for clear failure messages) ──────────

test_that("docs block 'minimal_workflow' is self-contained and correct", {
  md_file <- file.path(dirname(dirname(getwd())), "docs", "user_guide.md")
  skip_if_not(file.exists(md_file), "docs/user_guide.md not found")
  blocks <- .extract_r_docs_blocks(md_file)
  skip_if_not("minimal_workflow" %in% names(blocks))
  env <- new.env(parent = globalenv())
  expect_no_error(eval(parse(text = blocks[["minimal_workflow"]]), envir = env))
  expect_true(exists("fit", envir = env))
  expect_s3_class(env$fit, "werm_fit")
})

test_that("docs block 'predict_logistic' returns correct shapes", {
  md_file <- file.path(dirname(dirname(getwd())), "docs", "user_guide.md")
  skip_if_not(file.exists(md_file), "docs/user_guide.md not found")
  blocks <- .extract_r_docs_blocks(md_file)
  skip_if_not("predict_logistic" %in% names(blocks))
  env <- new.env(parent = globalenv())
  expect_no_error(eval(parse(text = blocks[["predict_logistic"]]), envir = env))
  expect_equal(length(env$labels), 10L)
  expect_equal(dim(env$probs), c(10L, 2L))
})

# ---------------------------------------------------------------------------
# CV estimators
# ---------------------------------------------------------------------------

#' Fit a WERM changepoint model with squared loss and CV selection
#'
#' Selects the number of changepoints from `{0, …, max_num_chgpts}` by
#' interleaved cross-validation, then refits on the full data.
#'
#' @param X Numeric matrix of shape `(n, p)`. Observations must be in order.
#' @param y Numeric vector of length `n`. Ordered responses.
#' @param max_num_chgpts Integer. Upper bound of the CV search grid.
#' @param delta Integer. Minimum gap between candidate changepoints.
#'   Default `1`.
#' @param search_method Character. `"efficient"` (default) or `"brute_force"`.
#' @param cv Integer. Number of interleaved folds. Default `5`.
#' @param fit_intercept Logical. Default `TRUE`.
#' @param use_base_loss_for_cv Logical. If `TRUE`, uses squared loss for CV
#'   segment fits and scoring instead of the default absolute error.
#'   Default `FALSE`.
#' @param penalty Character. Passed to the inner fixed model. Default `"none"`.
#' @param alpha Numeric. Penalty strength. Default `0`.
#'
#' @return An object of class `c("werm_least_squares_cv", "werm_fit")` with all
#'   elements from [werm_least_squares()] plus:
#'   \describe{
#'     \item{`best_num_chgpts`}{Integer. CV-selected number of changepoints.}
#'     \item{`best_score`}{Numeric. Mean held-out score for the best model.}
#'     \item{`cv_results`}{Data frame with columns `num_chgpts` and
#'       `mean_test_score`.}
#'     \item{`num_chgpts_grid`}{Integer vector. Full CV search grid.}
#'     \item{`segment_bounds`}{List of integer pairs `[start, stop]`
#'       (1-indexed, half-open).}
#'     \item{`segment_coefs`}{Matrix `(num_signals x p)`.}
#'     \item{`segment_intercepts`}{Numeric vector or `NULL`.}
#'   }
#'
#' @examples
#' # Limit BLAS/OpenMP threads so example CPU time stays proportional to
#' # elapsed time on multicore CRAN machines.
#' Sys.setenv(
#'   OMP_NUM_THREADS = "1",
#'   OPENBLAS_NUM_THREADS = "1",
#'   MKL_NUM_THREADS = "1",
#'   BLAS_NUM_THREADS = "1"
#' )
#'
#' if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
#'     weightederm_examples_available("WERMLeastSquaresCV")) {
#'   set.seed(10)
#'   n <- 30L; p <- 2L
#'   X <- matrix(rnorm(n * p), n, p)
#'   y <- c(
#'     X[1:10, ] %*% c(3.5, 0),
#'     X[11:20, ] %*% c(-3.5, 0),
#'     X[21:30, ] %*% c(3.5, 0)
#'   ) + rnorm(n, sd = 0.05)
#'   fit <- werm_least_squares_cv(X, y, max_num_chgpts = 2L, cv = 3L,
#'                                delta = 3L, fit_intercept = FALSE)
#'   fit$best_num_chgpts
#'   fit$changepoints
#'   fit$cv_results
#' }
#'
#' @export
werm_least_squares_cv <- function(
    X, y,
    max_num_chgpts,
    delta                = 1L,
    search_method        = "efficient",
    cv                   = 5L,
    fit_intercept        = TRUE,
    use_base_loss_for_cv = FALSE,
    penalty              = "none",
    alpha                = 0.0
) {
  py <- .weightederm_py()
  X_py <- .to_numpy_matrix(X)
  y_py <- .to_numpy_vector(y)

  py_model <- py$WERMLeastSquaresCV(
    max_num_chgpts       = as.integer(max_num_chgpts),
    delta                = as.integer(delta),
    search_method        = search_method,
    cv                   = as.integer(cv),
    fit_intercept        = fit_intercept,
    use_base_loss_for_cv = use_base_loss_for_cv,
    penalty              = penalty,
    alpha                = as.numeric(alpha)
  )
  py_model$fit(X_py, y_py)

  attrs <- .extract_cv_attrs(py_model)
  .build_werm_object(attrs, "least_squares_cv", py_model)
}


#' Fit a WERM changepoint model with Huber loss and CV selection
#'
#' @inheritParams werm_least_squares_cv
#' @param epsilon Numeric. Huber transition parameter. Default `1.35`.
#' @param max_iter Integer. Maximum L-BFGS-B iterations. Default `100`.
#' @param tol Numeric. Gradient-norm tolerance. Default `1e-5`.
#'
#' @return An object of class `c("werm_huber_cv", "werm_fit")`.
#'
#' @examples
#' # Limit BLAS/OpenMP threads so example CPU time stays proportional to
#' # elapsed time on multicore CRAN machines.
#' Sys.setenv(
#'   OMP_NUM_THREADS = "1",
#'   OPENBLAS_NUM_THREADS = "1",
#'   MKL_NUM_THREADS = "1",
#'   BLAS_NUM_THREADS = "1"
#' )
#'
#' if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
#'     weightederm_examples_available("WERMHuberCV")) {
#'   set.seed(11)
#'   n <- 24L; p <- 2L
#'   X <- matrix(rnorm(n * p), n, p)
#'   y <- c(
#'     X[1:8, ] %*% c(4, 0),
#'     X[9:16, ] %*% c(-4, 0),
#'     X[17:24, ] %*% c(4, 0)
#'   ) + rnorm(n, sd = 0.02)
#'   fit <- werm_huber_cv(X, y, max_num_chgpts = 2L, cv = 3L, delta = 2L,
#'                        fit_intercept = FALSE, max_iter = 20L)
#'   fit$best_num_chgpts
#' }
#'
#' @export
werm_huber_cv <- function(
    X, y,
    max_num_chgpts,
    delta                = 1L,
    search_method        = "efficient",
    cv                   = 5L,
    fit_intercept        = TRUE,
    epsilon              = 1.35,
    max_iter             = 100L,
    tol                  = 1e-5,
    use_base_loss_for_cv = FALSE,
    penalty              = "none",
    alpha                = 0.0
) {
  py <- .weightederm_py()
  X_py <- .to_numpy_matrix(X)
  y_py <- .to_numpy_vector(y)

  py_model <- py$WERMHuberCV(
    max_num_chgpts       = as.integer(max_num_chgpts),
    delta                = as.integer(delta),
    search_method        = search_method,
    cv                   = as.integer(cv),
    fit_intercept        = fit_intercept,
    epsilon              = as.numeric(epsilon),
    max_iter             = as.integer(max_iter),
    tol                  = as.numeric(tol),
    use_base_loss_for_cv = use_base_loss_for_cv,
    penalty              = penalty,
    alpha                = as.numeric(alpha)
  )
  py_model$fit(X_py, y_py)

  attrs <- .extract_cv_attrs(py_model)
  .build_werm_object(attrs, "huber_cv", py_model)
}


#' Fit a WERM changepoint model with logistic loss and CV selection
#'
#' @inheritParams werm_least_squares_cv
#' @param y Integer or factor vector of binary labels.
#' @param max_iter Integer. Maximum L-BFGS-B iterations. Default `100`.
#' @param tol Numeric. Gradient-norm tolerance. Default `1e-5`.
#' @param use_base_loss_for_cv Logical. If `TRUE`, uses logistic loss for CV
#'   segment scoring instead of the default absolute error. Default `FALSE`.
#' @param penalty Character. Default `"l2"`.
#' @param alpha Numeric. Default `1.0`.
#'
#' @return An object of class `c("werm_logistic_cv", "werm_fit")`.
#'   Contains all CV elements plus `classes` (character vector of length 2).
#'
#' @examples
#' # Limit BLAS/OpenMP threads so example CPU time stays proportional to
#' # elapsed time on multicore CRAN machines.
#' Sys.setenv(
#'   OMP_NUM_THREADS = "1",
#'   OPENBLAS_NUM_THREADS = "1",
#'   MKL_NUM_THREADS = "1",
#'   BLAS_NUM_THREADS = "1"
#' )
#'
#' if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
#'     weightederm_examples_available("WERMLogisticCV")) {
#'   set.seed(12)
#'   n <- 30L; p <- 2L
#'   X <- matrix(rnorm(n * p), n, p)
#'   eta <- c(
#'     X[1:10, ] %*% c(3, -3),
#'     X[11:20, ] %*% c(-3, 3),
#'     X[21:30, ] %*% c(3, -3)
#'   )
#'   y <- rbinom(n, 1L, 1 / (1 + exp(-eta)))
#'   fit <- werm_logistic_cv(X, y, max_num_chgpts = 2L, cv = 3L, delta = 3L,
#'                           fit_intercept = FALSE, max_iter = 100L)
#'   fit$best_num_chgpts
#'   fit$changepoints
#' }
#'
#' @export
werm_logistic_cv <- function(
    X, y,
    max_num_chgpts,
    delta                = 1L,
    search_method        = "efficient",
    cv                   = 5L,
    fit_intercept        = TRUE,
    max_iter             = 100L,
    tol                  = 1e-5,
    use_base_loss_for_cv = FALSE,
    penalty              = "l2",
    alpha                = 1.0
) {
  py <- .weightederm_py()
  X_py <- .to_numpy_matrix(X)
  y_py <- .to_numpy_vector(y)

  py_model <- py$WERMLogisticCV(
    max_num_chgpts       = as.integer(max_num_chgpts),
    delta                = as.integer(delta),
    search_method        = search_method,
    cv                   = as.integer(cv),
    fit_intercept        = fit_intercept,
    max_iter             = as.integer(max_iter),
    tol                  = as.numeric(tol),
    use_base_loss_for_cv = use_base_loss_for_cv,
    penalty              = penalty,
    alpha                = as.numeric(alpha)
  )
  py_model$fit(X_py, y_py)

  attrs <- .extract_cv_attrs(py_model)
  attrs$classes <- as.character(reticulate::py_to_r(py_model$classes_))

  .build_werm_object(attrs, "logistic_cv", py_model)
}

skip_if_not(weightederm_available, "weightederm Python package not available")

make_two_cp_data <- function(seed = 10L, n = 90L, p = 2L) {
  set.seed(seed)
  X <- matrix(rnorm(n * p), n, p)
  y <- c(
    X[1:30,  ] %*% c(3, 0)  + rnorm(30,  sd = 0.05),
    X[31:60, ] %*% c(-3, 0) + rnorm(30,  sd = 0.05),
    X[61:90, ] %*% c(3, 0)  + rnorm(30,  sd = 0.05)
  )
  list(X = X, y = y, true_cps = c(30L, 60L), n = n, p = p)
}

## ── WERMLeastSquaresCV ───────────────────────────────────────────────────────

test_that("werm_least_squares_cv selects correct num_chgpts", {
  d <- make_two_cp_data()
  fit <- werm_least_squares_cv(d$X, d$y, max_num_chgpts = 4L, cv = 3L,
                                delta = 5L, fit_intercept = FALSE)

  expect_equal(fit$best_num_chgpts, 2L)
})

test_that("cv_results has correct structure", {
  d <- make_two_cp_data()
  fit <- werm_least_squares_cv(d$X, d$y, max_num_chgpts = 3L, cv = 3L,
                                delta = 5L, fit_intercept = FALSE)

  expect_s3_class(fit$cv_results, "data.frame")
  expect_named(fit$cv_results, c("num_chgpts", "mean_test_score"))
  expect_equal(nrow(fit$cv_results), 4L)  # 0:3 inclusive
})

test_that("segment_bounds are 1-indexed and cover the full range", {
  d <- make_two_cp_data()
  fit <- werm_least_squares_cv(d$X, d$y, max_num_chgpts = 3L, cv = 3L,
                                delta = 5L, fit_intercept = FALSE)

  bounds <- fit$segment_bounds
  expect_equal(bounds[[1]][[1]], 1L)              # first segment starts at 1
  expect_equal(bounds[[length(bounds)]][[2]], d$n) # last segment ends at n
})

test_that("changepoints are close to true values", {
  d <- make_two_cp_data()
  fit <- werm_least_squares_cv(d$X, d$y, max_num_chgpts = 4L, cv = 3L,
                                delta = 5L, fit_intercept = FALSE)

  expect_equal(fit$num_chgpts, 2L)
  expect_lte(abs(fit$changepoints[1] - d$true_cps[1]), 5L)
  expect_lte(abs(fit$changepoints[2] - d$true_cps[2]), 5L)
})

test_that("predict works on CV estimator", {
  d <- make_two_cp_data()
  fit <- werm_least_squares_cv(d$X, d$y, max_num_chgpts = 3L, cv = 3L,
                                delta = 5L, fit_intercept = FALSE)
  X_new <- matrix(rnorm(8L * d$p), 8L, d$p)
  preds <- predict(fit, X_new)

  expect_length(preds, 8L)
  expect_true(all(is.finite(preds)))
})

## ── WERMHuberCV smoke test ───────────────────────────────────────────────────

test_that("werm_huber_cv runs and returns werm_fit", {
  d <- make_two_cp_data()
  fit <- werm_huber_cv(d$X, d$y, max_num_chgpts = 3L, cv = 3L,
                       delta = 5L, fit_intercept = FALSE)

  expect_s3_class(fit, "werm_fit")
  expect_true(fit$best_num_chgpts >= 0L)
})

## ── WERMLogisticCV smoke test ────────────────────────────────────────────────

test_that("werm_logistic_cv runs and returns werm_fit with classes", {
  set.seed(20L)
  n <- 80L; p <- 3L; true_cp <- 40L
  X  <- matrix(rnorm(n * p), n, p)
  eta <- c(X[1:true_cp,     ] %*% c(2, -1, 0),
           X[(true_cp+1):n, ] %*% c(-2, 1, 0))
  y  <- rbinom(n, 1L, 1 / (1 + exp(-eta)))

  fit <- werm_logistic_cv(X, y, max_num_chgpts = 2L, cv = 3L,
                           delta = 5L, fit_intercept = FALSE)

  expect_s3_class(fit, "werm_fit")
  expect_length(fit$classes, 2L)
  expect_true(fit$best_num_chgpts >= 0L)
})

# ---------------------------------------------------------------------------
# Input coercion
# ---------------------------------------------------------------------------

#' @noRd
.to_numpy_matrix <- function(X) {
  reticulate::r_to_py(as.matrix(X))
}

#' @noRd
.to_numpy_vector <- function(y) {
  reticulate::r_to_py(as.numeric(y))
}

#' @noRd
.weightederm_module_has_attrs <- function(module, attrs) {
  attrs <- as.character(attrs)

  if (length(attrs) == 0L) {
    return(TRUE)
  }

  py_attrs_present <- all(vapply(
    attrs,
    function(attr) {
      tryCatch(
        reticulate::py_has_attr(module, attr),
        error = function(e) FALSE
      )
    },
    logical(1)
  ))

  if (py_attrs_present) {
    return(TRUE)
  }

  if (is.environment(module)) {
    return(all(vapply(
      attrs,
      exists,
      logical(1),
      where = module,
      inherits = FALSE
    )))
  }

  if (is.list(module)) {
    return(all(attrs %in% names(module)))
  }

  FALSE
}

#' @noRd
.weightederm_examples_available <- function(required_attrs = character(),
                                            module = NULL) {
  if (is.null(module)) {
    if (!reticulate::py_module_available("weightederm")) {
      return(FALSE)
    }

    module <- tryCatch(
      reticulate::import("weightederm", delay_load = FALSE),
      error = function(e) NULL
    )

    if (is.null(module)) {
      return(FALSE)
    }
  }

  .weightederm_module_has_attrs(module, required_attrs)
}

#' Check whether the Python `weightederm` module is available for examples
#'
#' Returns `TRUE` only when the Python `weightederm` module can be imported and
#' exposes each requested attribute. This is primarily intended for guarding
#' package examples on systems where Python or the Python package is unavailable.
#'
#' @param required_attrs Character vector of Python attributes that must exist.
#'   Default `character()`.
#'
#' @return Logical scalar.
#' @keywords internal
#' @export
weightederm_examples_available <- function(required_attrs = character()) {
  .weightederm_examples_available(required_attrs = required_attrs)
}

# ---------------------------------------------------------------------------
# Extracting fitted attributes from a Python estimator
# ---------------------------------------------------------------------------

#' @noRd
.py_to_r_nullable <- function(x) {
  r_val <- reticulate::py_to_r(x)
  if (is.null(r_val)) NULL else r_val
}

#' @noRd
.extract_fixed_attrs <- function(py_model) {
  cp_0idx       <- as.integer(reticulate::py_to_r(py_model$changepoints_))
  coef          <- as.numeric(reticulate::py_to_r(py_model$last_segment_coef_))
  intercept_raw <- reticulate::py_to_r(py_model$last_segment_intercept_)

  list(
    changepoints           = cp_0idx + 1L,          # convert to 1-indexed
    num_chgpts             = as.integer(reticulate::py_to_r(py_model$num_chgpts_)),
    num_signals            = as.integer(reticulate::py_to_r(py_model$num_signals_)),
    objective              = as.numeric(reticulate::py_to_r(py_model$objective_)),
    n_features_in          = as.integer(reticulate::py_to_r(py_model$n_features_in_)),
    last_segment_coef      = coef,
    last_segment_intercept = if (is.null(intercept_raw)) NULL
                             else as.numeric(intercept_raw),
    signal_coefs           = reticulate::py_to_r(py_model$`_signal_coefs_`),
    signal_intercepts      = .py_to_r_nullable(py_model$`_signal_intercepts_`),
    theta_hat              = reticulate::py_to_r(py_model$`_theta_hat_`)
  )
}

#' @noRd
.extract_cv_attrs <- function(py_model) {
  base <- .extract_fixed_attrs(py_model)

  bounds_raw <- reticulate::py_to_r(py_model$segment_bounds_)
  # bounds_raw is a list of [start, stop] pairs (0-indexed) — convert to 1-indexed
  bounds_1idx <- lapply(bounds_raw, function(b) c(b[[1]] + 1L, b[[2]]))

  cv_results_raw <- reticulate::py_to_r(py_model$cv_results_)

  c(
    base,
    list(
      best_num_chgpts   = as.integer(reticulate::py_to_r(py_model$best_num_chgpts_)),
      best_index        = as.integer(reticulate::py_to_r(py_model$best_index_)) + 1L,
      best_score        = as.numeric(reticulate::py_to_r(py_model$best_score_)),
      cv_results        = data.frame(
        num_chgpts      = as.integer(cv_results_raw$num_chgpts),
        mean_test_score = as.numeric(cv_results_raw$mean_test_score)
      ),
      num_chgpts_grid   = as.integer(reticulate::py_to_r(py_model$num_chgpts_grid_)),
      segment_bounds    = bounds_1idx,
      segment_coefs     = reticulate::py_to_r(py_model$segment_coefs_),
      segment_intercepts = .py_to_r_nullable(py_model$segment_intercepts_)
    )
  )
}

#' @noRd
.build_werm_object <- function(attrs, estimator_type, py_model) {
  structure(
    c(attrs, list(
      .estimator_type = estimator_type,
      .py_model       = py_model       # keep reference for predict()
    )),
    class = c(paste0("werm_", estimator_type), "werm_fit")
  )
}

test_that("CRAN submission requirements stay satisfied", {
  source_root <- test_path("..", "..")
  skip_if_not(
    file.exists(file.path(source_root, "DESCRIPTION")),
    "source DESCRIPTION not available"
  )
  skip_if_not(
    dir.exists(file.path(source_root, "man")),
    "source man directory not available"
  )

  desc_dcf <- read.dcf(file.path(source_root, "DESCRIPTION"))
  description_field <- desc_dcf[1, "Description"]

  expect_match(description_field, "'weightederm'")
  expect_match(description_field, "'scikit-learn'")
  expect_match(description_field, "'reticulate'")
  expect_false(grepl("\\bWERM\\b", description_field))
  expect_false(grepl("\\bchangepoints\\b", description_field))
  # CRAN requires "authors (year) <doi:...>" format in Description.
  expect_match(description_field, "\\(2026\\)")
  expect_match(
    description_field,
    "<doi:10.48550/arXiv.2604.11746>",
    fixed = TRUE
  )

  man_dir <- file.path(source_root, "man")
  rd_files <- list.files(man_dir, pattern = "\\.Rd$", full.names = TRUE)
  rd_text <- setNames(
    vapply(
      rd_files,
      function(path) paste(readLines(path, warn = FALSE), collapse = "\n"),
      character(1)
    ),
    basename(rd_files)
  )

  # All exported topics should avoid \dontrun{} once an example can be made
  # executable in under 5 seconds.
  expect_false(any(grepl("\\\\dontrun\\s*\\{", rd_text)))

  # All werm_* estimator examples must avoid both wrappers and instead use the
  # if (.weightederm_examples_available()) guard.
  werm_rd <- rd_text[grep("^werm_", names(rd_text))]
  expect_false(any(grepl("\\\\donttest\\s*\\{", werm_rd)))

  # CRAN's Debian pretest flagged multithreaded example CPU time for werm_huber.
  # Keep thread limiting in all estimator examples so CPU time stays close to
  # elapsed time on multicore check machines.
  for (topic in names(werm_rd)) {
    expect_match(werm_rd[[topic]], "OMP_NUM_THREADS", info = topic)
    expect_match(werm_rd[[topic]], "OPENBLAS_NUM_THREADS", info = topic)
    expect_match(werm_rd[[topic]], "MKL_NUM_THREADS", info = topic)
    expect_match(werm_rd[[topic]], "BLAS_NUM_THREADS", info = topic)
    expect_match(werm_rd[[topic]], "RETICULATE_PYTHON", info = topic)
  }

  expect_match(werm_rd[["werm_least_squares.Rd"]], "n <- 24L", fixed = TRUE)
  expect_match(werm_rd[["werm_least_squares.Rd"]], "true_cp <- 12L", fixed = TRUE)
  expect_match(werm_rd[["werm_huber.Rd"]], "n <- 12L", fixed = TRUE)
  expect_match(werm_rd[["werm_huber.Rd"]], "p <- 1L", fixed = TRUE)
  expect_match(werm_rd[["werm_huber.Rd"]], "true_cp <- 6L", fixed = TRUE)
  expect_match(werm_rd[["werm_huber.Rd"]], "delta = 1L", fixed = TRUE)
  expect_match(werm_rd[["werm_huber.Rd"]], "max_iter = 5L", fixed = TRUE)
  expect_match(werm_rd[["werm_huber.Rd"]], "tol = 1e-3", fixed = TRUE)
  expect_match(werm_rd[["werm_logistic.Rd"]], "n <- 30L", fixed = TRUE)
  expect_match(werm_rd[["werm_logistic.Rd"]], "true_cp <- 15L", fixed = TRUE)
  expect_match(werm_rd[["werm_least_squares_cv.Rd"]], "n <- 30L", fixed = TRUE)
  expect_match(werm_rd[["werm_huber_cv.Rd"]], "n <- 24L", fixed = TRUE)
  expect_match(werm_rd[["werm_logistic_cv.Rd"]], "n <- 30L", fixed = TRUE)

  configure_python_rd <- rd_text[["weightederm_configure_python.Rd"]]
  expect_match(configure_python_rd, "Sys\\.which\\(\"python\"\\)")
  expect_match(configure_python_rd, "required = FALSE", fixed = TRUE)

  exported_topics <- c(
    "weightederm_configure_python.Rd",
    "werm_least_squares.Rd",
    "werm_huber.Rd",
    "werm_logistic.Rd",
    "werm_least_squares_cv.Rd",
    "werm_huber_cv.Rd",
    "werm_logistic_cv.Rd",
    "predict.werm_fit.Rd",
    "coef.werm_fit.Rd",
    "summary.werm_fit.Rd"
  )

  for (topic in exported_topics) {
    expect_match(
      rd_text[[topic]],
      "\\\\value\\s*\\{",
      info = topic
    )
  }
})

# ---------------------------------------------------------------------------
# print
# ---------------------------------------------------------------------------

#' @export
print.werm_fit <- function(x, ...) {
  is_cv <- inherits(x, c("werm_least_squares_cv", "werm_huber_cv",
                          "werm_logistic_cv"))
  loss_label <- switch(
    x$.estimator_type,
    least_squares    = "Least Squares",
    least_squares_cv = "Least Squares (CV)",
    huber            = "Huber",
    huber_cv         = "Huber (CV)",
    logistic         = "Logistic",
    logistic_cv      = "Logistic (CV)",
    x$.estimator_type
  )

  cat("WERM Changepoint Estimator ", loss_label, "\n")
  cat(sprintf("  Changepoints : %s (1-indexed)\n",
              if (length(x$changepoints) == 0) "none"
              else paste(x$changepoints, collapse = ", ")))
  cat(sprintf("  num_chgpts   : %d\n", x$num_chgpts))
  cat(sprintf("  num_signals  : %d\n", x$num_signals))
  cat(sprintf("  objective    : %.6g\n", x$objective))

  if (is_cv) {
    cat(sprintf("  best_score   : %.6g\n", x$best_score))
    cat("  cv_results   :\n")
    print(x$cv_results, row.names = FALSE)
  }

  invisible(x)
}


# ---------------------------------------------------------------------------
# predict
# ---------------------------------------------------------------------------

#' Predict using the last-segment refit
#'
#' Evaluates the unweighted base-loss model fitted on the last detected
#' segment. For logistic estimators, returns class labels.
#'
#' @param object A `werm_fit` object returned by any `werm_*` function.
#' @param newdata Numeric matrix of shape `(m, p)`.
#' @param type Character. For logistic estimators: `"class"` (default) or
#'   `"prob"` (returns a `(m, 2)` probability matrix).
#' @param ... Ignored.
#'
#' @return Numeric vector of length `m` (regression) or character vector /
#'   probability matrix (logistic).
#'
#' @export
predict.werm_fit <- function(object, newdata, type = "class", ...) {
  X_py <- .to_numpy_matrix(newdata)
  is_logistic <- inherits(object, c("werm_logistic", "werm_logistic_cv"))

  if (is_logistic && type == "prob") {
    raw <- reticulate::py_to_r(object$.py_model$predict_proba(X_py))
    return(raw)
  }

  raw <- reticulate::py_to_r(object$.py_model$predict(X_py))

  if (is_logistic) {
    as.character(raw)
  } else {
    as.numeric(raw)
  }
}


# ---------------------------------------------------------------------------
# coef
# ---------------------------------------------------------------------------

#' Extract last-segment coefficients
#'
#' Returns the coefficient vector of the unweighted base-loss refit on the
#' last detected segment (the same coefficients used by `predict()`).
#'
#' @param object A `werm_fit` object.
#' @param ... Ignored.
#'
#' @return Named numeric vector of length `p`.
#'
#' @export
coef.werm_fit <- function(object, ...) {
  object$last_segment_coef
}


# ---------------------------------------------------------------------------
# summary
# ---------------------------------------------------------------------------

#' Summarise a fitted WERM model
#'
#' @param object A `werm_fit` object.
#' @param ... Ignored.
#'
#' @return Invisibly returns `object`. Called for its printed side-effect.
#'
#' @export
summary.werm_fit <- function(object, ...) {
  print(object)

  cat("\nLast-segment coefficient(s):\n")
  print(object$last_segment_coef)

  if (!is.null(object$last_segment_intercept)) {
    cat(sprintf("Last-segment intercept: %.6g\n", object$last_segment_intercept))
  }

  cat("\nStage-1 signal coefficients (WERM-weighted, one row per signal):\n")
  print(object$signal_coefs)

  invisible(object)
}

## Limit numpy/BLAS threading so that CPU time stays proportional to
## elapsed time on multi-core CRAN check machines (avoids the
## "CPU time > 2.5 times elapsed" NOTE on Debian).
Sys.setenv(
  OMP_NUM_THREADS    = "1",
  OPENBLAS_NUM_THREADS = "1",
  MKL_NUM_THREADS    = "1",
  BLAS_NUM_THREADS   = "1"
)

## Point reticulate at the project venv that has weightederm installed.
## Skip all tests if Python / weightederm is unavailable.
## During testthat execution, getwd() is <pkg>/tests/testthat/.
## We need to go up 3 levels: tests/testthat/ → tests/ → weightederm-r/ → Weighted-ERM packages/
PYTHON_PATH <- file.path(
  dirname(dirname(dirname(getwd()))),
  "weightederm", ".venv", "bin", "python"
)

if (file.exists(PYTHON_PATH)) {
  reticulate::use_python(PYTHON_PATH, required = TRUE)
} else {
  # Fall back: let reticulate find whatever Python it can
}

weightederm_available <- tryCatch({
  weightederm::weightederm_examples_available(
    c(
      "WERMLeastSquares",
      "WERMLeastSquaresCV",
      "WERMHuber",
      "WERMHuberCV",
      "WERMLogistic",
      "WERMLogisticCV"
    )
  )
}, error = function(e) FALSE)

if (!weightederm_available) {
  message("compatible weightederm Python package not found — skipping all tests.")
}

test_that("example availability requires the requested Python exports", {
  module_has_attrs <- getFromNamespace(".weightederm_module_has_attrs", "weightederm")
  examples_available <- getFromNamespace(".weightederm_examples_available", "weightederm")
  module <- new.env(parent = emptyenv())
  module$WERMLeastSquares <- TRUE
  module$WERMLogistic <- TRUE

  expect_true(
    module_has_attrs(module, "WERMLeastSquares")
  )
  expect_false(
    module_has_attrs(
      module,
      c("WERMLeastSquares", "WERMHuber")
    )
  )
  expect_true(
    examples_available(
      "WERMLeastSquares",
      module = module
    )
  )
  expect_false(
    examples_available(
      "WERMHuber",
      module = module
    )
  )
})

test_that("example availability recognises attributes on a live Python module", {
  module_has_attrs <- getFromNamespace(".weightederm_module_has_attrs", "weightederm")
  examples_available <- getFromNamespace(".weightederm_examples_available", "weightederm")
  source_root <- normalizePath(test_path("..", ".."), winslash = "/", mustWork = TRUE)
  python_path <- file.path(
    dirname(source_root),
    "weightederm", ".venv", "bin", "python"
  )
  skip_if_not(file.exists(python_path), "project Python virtualenv not found")

  reticulate::use_python(python_path, required = TRUE)
  skip_if_not(
    reticulate::py_module_available("weightederm"),
    "weightederm Python package not available"
  )

  module <- reticulate::import("weightederm", delay_load = FALSE)

  expect_true(
    module_has_attrs(
      module,
      c("WERMLeastSquares", "WERMHuber", "WERMLogistic")
    )
  )
  expect_true(
    examples_available(
      c("WERMLeastSquares", "WERMHuber", "WERMLogistic"),
      module = module
    )
  )
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimators.R
\name{werm_least_squares}
\alias{werm_least_squares}
\title{Fit a WERM changepoint model with squared loss}
\usage{
werm_least_squares(
  X,
  y,
  num_chgpts,
  delta = 1L,
  search_method = "efficient",
  fit_intercept = TRUE,
  fit_solver = "direct",
  penalty = "none",
  alpha = 0
)
}
\arguments{
\item{X}{Numeric matrix of shape \verb{(n, p)}. Observations must be in order.}

\item{y}{Numeric vector of length \code{n}. Ordered responses.}

\item{num_chgpts}{Integer. Number of changepoints to detect.}

\item{delta}{Integer. Minimum gap between candidate changepoints during
search. Default \code{1}. Rule of thumb: \code{max(1L, as.integer(0.05 * nrow(X)))}.}

\item{search_method}{Character. \code{"efficient"} (default) or \code{"brute_force"}.}

\item{fit_intercept}{Logical. Whether each segment model includes an
intercept. Default \code{TRUE}.}

\item{fit_solver}{Character. \code{"direct"} (default, closed-form) or
\code{"lbfgsb"} (gradient-based).}

\item{penalty}{Character. \code{"none"} (default), \code{"l1"}, or \code{"l2"}.}

\item{alpha}{Numeric. Penalty strength. Default \code{0}.}
}
\value{
An object of class \code{c("werm_least_squares", "werm_fit")} with the
following named elements:
\describe{
\item{\code{changepoints}}{Integer vector of detected changepoints
(\strong{1-indexed}, R convention). Length equals \code{num_chgpts}.}
\item{\code{num_chgpts}}{Integer. Number of detected changepoints.}
\item{\code{num_signals}}{Integer. Number of segments (\code{num_chgpts + 1}).}
\item{\code{objective}}{Numeric. Minimised WERM objective value.}
\item{\code{last_segment_coef}}{Numeric vector. Coefficient of the
unweighted base-loss refit on the last segment (used by \code{predict()}).}
\item{\code{last_segment_intercept}}{Numeric or \code{NULL}.}
\item{\code{signal_coefs}}{Matrix \verb{(num_signals x p)}. Stage-1 WERM
coefficient estimates.}
\item{\code{signal_intercepts}}{Numeric vector or \code{NULL}.}
\item{\code{n_features_in}}{Integer. Number of features.}
}
}
\description{
Detects \code{num_chgpts} changepoints in ordered regression data by minimising
a Weighted Empirical Risk with squared loss.
}
\examples{
# Limit BLAS/OpenMP threads so example CPU time stays proportional to
# elapsed time on multicore CRAN machines.
Sys.setenv(
  OMP_NUM_THREADS = "1",
  OPENBLAS_NUM_THREADS = "1",
  MKL_NUM_THREADS = "1",
  BLAS_NUM_THREADS = "1"
)

if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
    weightederm_examples_available("WERMLeastSquares")) {
  set.seed(1)
  n <- 24L; p <- 2L; true_cp <- 12L
  X <- matrix(rnorm(n * p), n, p)
  y <- c(
    X[1:true_cp, ] \%*\% c(3, -1.5),
    X[(true_cp + 1L):n, ] \%*\% c(-3, 1.5)
  ) + rnorm(n, sd = 0.05)
  fit <- werm_least_squares(X, y, num_chgpts = 1L, delta = 3L,
                            fit_intercept = FALSE)
  fit$changepoints
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{coef.werm_fit}
\alias{coef.werm_fit}
\title{Extract last-segment coefficients}
\usage{
\method{coef}{werm_fit}(object, ...)
}
\arguments{
\item{object}{A \code{werm_fit} object.}

\item{...}{Ignored.}
}
\value{
Named numeric vector of length \code{p}.
}
\description{
Returns the coefficient vector of the unweighted base-loss refit on the
last detected segment (the same coefficients used by \code{predict()}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{predict.werm_fit}
\alias{predict.werm_fit}
\title{Predict using the last-segment refit}
\usage{
\method{predict}{werm_fit}(object, newdata, type = "class", ...)
}
\arguments{
\item{object}{A \code{werm_fit} object returned by any \verb{werm_*} function.}

\item{newdata}{Numeric matrix of shape \verb{(m, p)}.}

\item{type}{Character. For logistic estimators: \code{"class"} (default) or
\code{"prob"} (returns a \verb{(m, 2)} probability matrix).}

\item{...}{Ignored.}
}
\value{
Numeric vector of length \code{m} (regression) or character vector /
probability matrix (logistic).
}
\description{
Evaluates the unweighted base-loss model fitted on the last detected
segment. For logistic estimators, returns class labels.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{weightederm_configure_python}
\alias{weightederm_configure_python}
\title{Configure which Python environment weightederm uses}
\usage{
weightederm_configure_python(python, required = TRUE)
}
\arguments{
\item{python}{Path to the Python binary or virtual-environment directory.}

\item{required}{Passed to \code{\link[reticulate:use_python]{reticulate::use_python()}}.}
}
\value{
Invisibly returns \code{NULL}. Called for side effects only.
}
\description{
A thin wrapper around \code{\link[reticulate:use_python]{reticulate::use_python()}} that must be called
\strong{before} any \verb{werm_*} function if the default Python does not have
\code{weightederm} installed.
}
\examples{
if (nzchar(Sys.which("python"))) {
  weightederm_configure_python(Sys.which("python"), required = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{weightederm_examples_available}
\alias{weightederm_examples_available}
\title{Check whether the Python \code{weightederm} module is available for examples}
\usage{
weightederm_examples_available(required_attrs = character())
}
\arguments{
\item{required_attrs}{Character vector of Python attributes that must exist.
Default \code{character()}.}
}
\value{
Logical scalar.
}
\description{
Returns \code{TRUE} only when the Python \code{weightederm} module can be imported and
exposes each requested attribute. This is primarily intended for guarding
package examples on systems where Python or the Python package is unavailable.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimators.R
\name{werm_logistic}
\alias{werm_logistic}
\title{Fit a WERM changepoint model with binary logistic loss}
\usage{
werm_logistic(
  X,
  y,
  num_chgpts,
  delta = 1L,
  search_method = "efficient",
  fit_intercept = TRUE,
  max_iter = 100L,
  tol = 1e-05,
  penalty = "l2",
  alpha = 1
)
}
\arguments{
\item{X}{Numeric matrix of shape \verb{(n, p)}. Observations must be in order.}

\item{y}{Integer or factor vector of binary labels (two unique values).}

\item{num_chgpts}{Integer. Number of changepoints to detect.}

\item{delta}{Integer. Minimum gap between candidate changepoints during
search. Default \code{1}. Rule of thumb: \code{max(1L, as.integer(0.05 * nrow(X)))}.}

\item{search_method}{Character. \code{"efficient"} (default) or \code{"brute_force"}.}

\item{fit_intercept}{Logical. Whether each segment model includes an
intercept. Default \code{TRUE}.}

\item{max_iter}{Integer. Maximum L-BFGS-B iterations. Default \code{100}.}

\item{tol}{Numeric. Gradient-norm tolerance. Default \code{1e-5}.}

\item{penalty}{Character. Default \code{"l2"} (prevents divergence on separable
segments). Set \code{"none"} only when segments are not separable.}

\item{alpha}{Numeric. Default \code{1.0}.}
}
\value{
An object of class \code{c("werm_logistic", "werm_fit")}.
Contains all elements from \code{\link[=werm_least_squares]{werm_least_squares()}} plus:
\describe{
\item{\code{classes}}{Character vector of length 2. The two class labels in
sorted order. \code{classes[2]} is the positive class.}
}
}
\description{
Detects \code{num_chgpts} changepoints in ordered binary classification data.
}
\examples{
# Limit BLAS/OpenMP threads so example CPU time stays proportional to
# elapsed time on multicore CRAN machines.
Sys.setenv(
  OMP_NUM_THREADS = "1",
  OPENBLAS_NUM_THREADS = "1",
  MKL_NUM_THREADS = "1",
  BLAS_NUM_THREADS = "1"
)

if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
    weightederm_examples_available("WERMLogistic")) {
  set.seed(3)
  n <- 30L; p <- 2L; true_cp <- 15L
  X <- matrix(rnorm(n * p), n, p)
  eta <- c(
    X[1:true_cp, ] \%*\% c(3, -3),
    X[(true_cp + 1L):n, ] \%*\% c(-3, 3)
  )
  y <- rbinom(n, 1L, 1 / (1 + exp(-eta)))
  fit <- werm_logistic(X, y, num_chgpts = 1L, delta = 3L,
                       fit_intercept = FALSE, max_iter = 100L)
  fit$changepoints
  fit$classes
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.werm_fit}
\alias{summary.werm_fit}
\title{Summarise a fitted WERM model}
\usage{
\method{summary}{werm_fit}(object, ...)
}
\arguments{
\item{object}{A \code{werm_fit} object.}

\item{...}{Ignored.}
}
\value{
Invisibly returns \code{object}. Called for its printed side-effect.
}
\description{
Summarise a fitted WERM model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cv_estimators.R
\name{werm_least_squares_cv}
\alias{werm_least_squares_cv}
\title{Fit a WERM changepoint model with squared loss and CV selection}
\usage{
werm_least_squares_cv(
  X,
  y,
  max_num_chgpts,
  delta = 1L,
  search_method = "efficient",
  cv = 5L,
  fit_intercept = TRUE,
  use_base_loss_for_cv = FALSE,
  penalty = "none",
  alpha = 0
)
}
\arguments{
\item{X}{Numeric matrix of shape \verb{(n, p)}. Observations must be in order.}

\item{y}{Numeric vector of length \code{n}. Ordered responses.}

\item{max_num_chgpts}{Integer. Upper bound of the CV search grid.}

\item{delta}{Integer. Minimum gap between candidate changepoints.
Default \code{1}.}

\item{search_method}{Character. \code{"efficient"} (default) or \code{"brute_force"}.}

\item{cv}{Integer. Number of interleaved folds. Default \code{5}.}

\item{fit_intercept}{Logical. Default \code{TRUE}.}

\item{use_base_loss_for_cv}{Logical. If \code{TRUE}, uses squared loss for CV
segment fits and scoring instead of the default absolute error.
Default \code{FALSE}.}

\item{penalty}{Character. Passed to the inner fixed model. Default \code{"none"}.}

\item{alpha}{Numeric. Penalty strength. Default \code{0}.}
}
\value{
An object of class \code{c("werm_least_squares_cv", "werm_fit")} with all
elements from \code{\link[=werm_least_squares]{werm_least_squares()}} plus:
\describe{
\item{\code{best_num_chgpts}}{Integer. CV-selected number of changepoints.}
\item{\code{best_score}}{Numeric. Mean held-out score for the best model.}
\item{\code{cv_results}}{Data frame with columns \code{num_chgpts} and
\code{mean_test_score}.}
\item{\code{num_chgpts_grid}}{Integer vector. Full CV search grid.}
\item{\code{segment_bounds}}{List of integer pairs \verb{[start, stop]}
(1-indexed, half-open).}
\item{\code{segment_coefs}}{Matrix \verb{(num_signals x p)}.}
\item{\code{segment_intercepts}}{Numeric vector or \code{NULL}.}
}
}
\description{
Selects the number of changepoints from \verb{\{0, …, max_num_chgpts\}} by
interleaved cross-validation, then refits on the full data.
}
\examples{
# Limit BLAS/OpenMP threads so example CPU time stays proportional to
# elapsed time on multicore CRAN machines.
Sys.setenv(
  OMP_NUM_THREADS = "1",
  OPENBLAS_NUM_THREADS = "1",
  MKL_NUM_THREADS = "1",
  BLAS_NUM_THREADS = "1"
)

if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
    weightederm_examples_available("WERMLeastSquaresCV")) {
  set.seed(10)
  n <- 30L; p <- 2L
  X <- matrix(rnorm(n * p), n, p)
  y <- c(
    X[1:10, ] \%*\% c(3.5, 0),
    X[11:20, ] \%*\% c(-3.5, 0),
    X[21:30, ] \%*\% c(3.5, 0)
  ) + rnorm(n, sd = 0.05)
  fit <- werm_least_squares_cv(X, y, max_num_chgpts = 2L, cv = 3L,
                               delta = 3L, fit_intercept = FALSE)
  fit$best_num_chgpts
  fit$changepoints
  fit$cv_results
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimators.R
\name{werm_huber}
\alias{werm_huber}
\title{Fit a WERM changepoint model with Huber loss}
\usage{
werm_huber(
  X,
  y,
  num_chgpts,
  delta = 1L,
  search_method = "efficient",
  fit_intercept = TRUE,
  epsilon = 1.35,
  max_iter = 100L,
  tol = 1e-05,
  penalty = "none",
  alpha = 0
)
}
\arguments{
\item{X}{Numeric matrix of shape \verb{(n, p)}. Observations must be in order.}

\item{y}{Numeric vector of length \code{n}. Ordered responses.}

\item{num_chgpts}{Integer. Number of changepoints to detect.}

\item{delta}{Integer. Minimum gap between candidate changepoints during
search. Default \code{1}. Rule of thumb: \code{max(1L, as.integer(0.05 * nrow(X)))}.}

\item{search_method}{Character. \code{"efficient"} (default) or \code{"brute_force"}.}

\item{fit_intercept}{Logical. Whether each segment model includes an
intercept. Default \code{TRUE}.}

\item{epsilon}{Numeric. Huber transition parameter. Default \code{1.35}
(95 \% Gaussian efficiency).}

\item{max_iter}{Integer. Maximum L-BFGS-B iterations. Default \code{100}.}

\item{tol}{Numeric. Gradient-norm tolerance. Default \code{1e-5}.}

\item{penalty}{Character. \code{"none"} (default), \code{"l1"}, or \code{"l2"}.}

\item{alpha}{Numeric. Penalty strength. Default \code{0}.}
}
\value{
An object of class \code{c("werm_huber", "werm_fit")}.
Same elements as \code{\link[=werm_least_squares]{werm_least_squares()}}.
}
\description{
Like \code{\link[=werm_least_squares]{werm_least_squares()}} but uses the Huber loss, which is more robust
to outliers than squared loss.
}
\examples{
# Limit BLAS/OpenMP threads so example CPU time stays proportional to
# elapsed time on multicore CRAN machines.
Sys.setenv(
  OMP_NUM_THREADS = "1",
  OPENBLAS_NUM_THREADS = "1",
  MKL_NUM_THREADS = "1",
  BLAS_NUM_THREADS = "1"
)

if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
    weightederm_examples_available("WERMHuber")) {
  set.seed(2)
  n <- 12L; p <- 1L; true_cp <- 6L
  X <- matrix(rnorm(n * p), n, p)
  y <- c(
    X[1:true_cp, , drop = FALSE] \%*\% 3,
    X[(true_cp + 1L):n, , drop = FALSE] \%*\% -3
  ) + rnorm(n, sd = 0.03)
  fit <- werm_huber(X, y, num_chgpts = 1L, delta = 1L,
                    fit_intercept = FALSE, max_iter = 5L, tol = 1e-3)
  fit$changepoints
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cv_estimators.R
\name{werm_huber_cv}
\alias{werm_huber_cv}
\title{Fit a WERM changepoint model with Huber loss and CV selection}
\usage{
werm_huber_cv(
  X,
  y,
  max_num_chgpts,
  delta = 1L,
  search_method = "efficient",
  cv = 5L,
  fit_intercept = TRUE,
  epsilon = 1.35,
  max_iter = 100L,
  tol = 1e-05,
  use_base_loss_for_cv = FALSE,
  penalty = "none",
  alpha = 0
)
}
\arguments{
\item{X}{Numeric matrix of shape \verb{(n, p)}. Observations must be in order.}

\item{y}{Numeric vector of length \code{n}. Ordered responses.}

\item{max_num_chgpts}{Integer. Upper bound of the CV search grid.}

\item{delta}{Integer. Minimum gap between candidate changepoints.
Default \code{1}.}

\item{search_method}{Character. \code{"efficient"} (default) or \code{"brute_force"}.}

\item{cv}{Integer. Number of interleaved folds. Default \code{5}.}

\item{fit_intercept}{Logical. Default \code{TRUE}.}

\item{epsilon}{Numeric. Huber transition parameter. Default \code{1.35}.}

\item{max_iter}{Integer. Maximum L-BFGS-B iterations. Default \code{100}.}

\item{tol}{Numeric. Gradient-norm tolerance. Default \code{1e-5}.}

\item{use_base_loss_for_cv}{Logical. If \code{TRUE}, uses squared loss for CV
segment fits and scoring instead of the default absolute error.
Default \code{FALSE}.}

\item{penalty}{Character. Passed to the inner fixed model. Default \code{"none"}.}

\item{alpha}{Numeric. Penalty strength. Default \code{0}.}
}
\value{
An object of class \code{c("werm_huber_cv", "werm_fit")}.
}
\description{
Fit a WERM changepoint model with Huber loss and CV selection
}
\examples{
# Limit BLAS/OpenMP threads so example CPU time stays proportional to
# elapsed time on multicore CRAN machines.
Sys.setenv(
  OMP_NUM_THREADS = "1",
  OPENBLAS_NUM_THREADS = "1",
  MKL_NUM_THREADS = "1",
  BLAS_NUM_THREADS = "1"
)

if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
    weightederm_examples_available("WERMHuberCV")) {
  set.seed(11)
  n <- 24L; p <- 2L
  X <- matrix(rnorm(n * p), n, p)
  y <- c(
    X[1:8, ] \%*\% c(4, 0),
    X[9:16, ] \%*\% c(-4, 0),
    X[17:24, ] \%*\% c(4, 0)
  ) + rnorm(n, sd = 0.02)
  fit <- werm_huber_cv(X, y, max_num_chgpts = 2L, cv = 3L, delta = 2L,
                       fit_intercept = FALSE, max_iter = 20L)
  fit$best_num_chgpts
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cv_estimators.R
\name{werm_logistic_cv}
\alias{werm_logistic_cv}
\title{Fit a WERM changepoint model with logistic loss and CV selection}
\usage{
werm_logistic_cv(
  X,
  y,
  max_num_chgpts,
  delta = 1L,
  search_method = "efficient",
  cv = 5L,
  fit_intercept = TRUE,
  max_iter = 100L,
  tol = 1e-05,
  use_base_loss_for_cv = FALSE,
  penalty = "l2",
  alpha = 1
)
}
\arguments{
\item{X}{Numeric matrix of shape \verb{(n, p)}. Observations must be in order.}

\item{y}{Integer or factor vector of binary labels.}

\item{max_num_chgpts}{Integer. Upper bound of the CV search grid.}

\item{delta}{Integer. Minimum gap between candidate changepoints.
Default \code{1}.}

\item{search_method}{Character. \code{"efficient"} (default) or \code{"brute_force"}.}

\item{cv}{Integer. Number of interleaved folds. Default \code{5}.}

\item{fit_intercept}{Logical. Default \code{TRUE}.}

\item{max_iter}{Integer. Maximum L-BFGS-B iterations. Default \code{100}.}

\item{tol}{Numeric. Gradient-norm tolerance. Default \code{1e-5}.}

\item{use_base_loss_for_cv}{Logical. If \code{TRUE}, uses logistic loss for CV
segment scoring instead of the default absolute error. Default \code{FALSE}.}

\item{penalty}{Character. Default \code{"l2"}.}

\item{alpha}{Numeric. Default \code{1.0}.}
}
\value{
An object of class \code{c("werm_logistic_cv", "werm_fit")}.
Contains all CV elements plus \code{classes} (character vector of length 2).
}
\description{
Fit a WERM changepoint model with logistic loss and CV selection
}
\examples{
# Limit BLAS/OpenMP threads so example CPU time stays proportional to
# elapsed time on multicore CRAN machines.
Sys.setenv(
  OMP_NUM_THREADS = "1",
  OPENBLAS_NUM_THREADS = "1",
  MKL_NUM_THREADS = "1",
  BLAS_NUM_THREADS = "1"
)

if (nzchar(Sys.getenv("RETICULATE_PYTHON")) &&
    weightederm_examples_available("WERMLogisticCV")) {
  set.seed(12)
  n <- 30L; p <- 2L
  X <- matrix(rnorm(n * p), n, p)
  eta <- c(
    X[1:10, ] \%*\% c(3, -3),
    X[11:20, ] \%*\% c(-3, 3),
    X[21:30, ] \%*\% c(3, -3)
  )
  y <- rbinom(n, 1L, 1 / (1 + exp(-eta)))
  fit <- werm_logistic_cv(X, y, max_num_chgpts = 2L, cv = 3L, delta = 3L,
                          fit_intercept = FALSE, max_iter = 100L)
  fit$best_num_chgpts
  fit$changepoints
}

}
